// Pre-packaged DDE Client class that encapsulates DDE and combines some functions
// See "ddemgr.h" for instructions for use
// Intellectual property of Lance Klein 11/93

#include "ddemgr.h"

char * DDEManager::GetDDEItem(char * DDEService, char * DDETopic, char * DDEItem,
						char * retdata/* = NULL */, int lenretdata/* = 0 */,
						HINSTANCE Inst/* = NULL */)
{
	// Perform all functions to start conversation and get one item of data
	if( Inst && !Initialize(Inst) )
		return NULL;  
	if( !Connect(DDEService, DDETopic) )
		return NULL;

	RetData = Request(DDEItem, retdata, lenretdata);

	Disconnect();

	return RetData;
}


BOOL DDEManager::Initialize(HINSTANCE Inst)
{
	if( idInst )
		return TRUE;

	hInst = Inst;
    idInst = 0;		// Must be 0 first call to DdeInitialize

	if( !(lpCallBack = MakeProcInstance( (FARPROC) DDEManager::CallBack, Inst)) )
	{
		strcpy(errortext, "MakeProcInstance");
		return FALSE;
	}

	if( DdeInitialize( &idInst, (PFNCALLBACK) lpCallBack, APPCMD_CLIENTONLY, 0) !=
		DMLERR_NO_ERROR )
	{
		strcpy(errortext, "DdeInitialize");
		return FALSE;
	}

	return TRUE;
}


BOOL DDEManager::Connect(char * DDEService, char * DDETopic,
									HINSTANCE Inst/* = NULL */)
{
	if( Inst && !Initialize(Inst) )
		return FALSE;

	// Must call Initialize first
	if( !idInst )
	{
		strcpy(errortext, "Not Initialized");
		return FALSE;
	}

    // Must Disconnect previous connection first
	if( hConv )
	{
		strcpy(errortext, "Already Connected");
		return FALSE;
	}

    // Get string handles for service and topic
	if( !(hszService = DdeCreateStringHandle(idInst, DDEService, CP_WINANSI)) )
	{
		strcpy(errortext, "DdeCreateStringHandle");
		return FALSE;
	}
	if( !(hszTopic = DdeCreateStringHandle(idInst, DDETopic, CP_WINANSI)) )
	{
		strcpy(errortext, "DdeCreateStringHandle");
		return FALSE;
	}

	// Set static "this" pointer for callback function
	SThis = this;

    // Connect to DDE Server
	if( !(hConv = DdeConnect(idInst, hszService, hszTopic, NULL)) )
	{
    	strcpy(errortext, "DdeConnect");
		return FALSE;
	}

	return TRUE;
}


char * DDEManager::Request(char * DDEItem, char * retdata/* = NULL */,
							int lenretdata/* = 0 */)
{
	if( !hConv )
	{
		strcpy(errortext, "Not Connected");
		return NULL;
	}

    strcpy(errortext, "");
	RetData = retdata;
	LenRetData = lenretdata;

    // Get string handle for item (free old one first)
	if( hszItem )
    {
		DdeFreeStringHandle(idInst, hszItem);
		hszItem = 0;
	}
	if( !(hszItem = DdeCreateStringHandle(idInst, DDEItem, CP_WINANSI)) )
	{
		strcpy(errortext, "DdeCreateStringHandle");
		return NULL;
	}

	// Set static "this" pointer for callback function
	SThis = this;

    // Initiate client transaction to get data item
	HDDEDATA hData = DdeClientTransaction(NULL, 0, hConv, hszItem, CF_TEXT,
						XTYP_REQUEST, 0, NULL);
	if( hData )
    {
		ReceivedData(hData);
		DdeFreeDataHandle(hData);
	}

	return RetData;
}


BOOL DDEManager::Poke(char * DDEItem, char * data)
{
	if( !hConv )
		return FALSE;

	SThis = this;

    // Get string handle for item
	if( hszItem )
    {
		DdeFreeStringHandle(idInst, hszItem);
		hszItem = 0;
	}
	if( !(hszItem = DdeCreateStringHandle(idInst, DDEItem, CP_WINANSI)) )
		return FALSE;

	// Send data
	return( (DdeClientTransaction((LPBYTE) data, strlen(data)+1, hConv,
    			hszItem, CF_TEXT, XTYP_POKE, 1000, NULL) != FALSE) );
}


// Execute a command
BOOL DDEManager::Execute(char * command)
{
	if( !hConv )
		return FALSE;

	SThis = this;

	// Send data
	return( (DdeClientTransaction((LPBYTE) command, strlen(command),
			hConv, NULL, CF_TEXT, XTYP_EXECUTE, 1000, NULL) != FALSE) );
}


BOOL DDEManager::Disconnect()
{
	if( !hConv )
		return FALSE;

	SThis = this;

    // Disconnect from server
	DdeDisconnect( hConv );
	hConv = 0;

    // Free string handles
	if( hszService )
	{
		DdeFreeStringHandle(idInst, hszService);
		hszService = 0;
	}
	if( hszTopic )
	{
		DdeFreeStringHandle(idInst, hszTopic);
		hszTopic = 0;
	}
	if( hszItem )
	{
		DdeFreeStringHandle(idInst, hszItem);
		hszItem = 0;
	}

	// Free callback function pointer
	if( lpCallBack )
	{
		FreeProcInstance((FARPROC) lpCallBack);
		lpCallBack = NULL;
    }

	return TRUE;
}


BOOL DDEManager::IsConnected()
{
	if( hConv )
		return TRUE;
    else
		return FALSE;
}


char * DDEManager::GetErrorText()
{
	return errortext;
}


void DDEManager::ReceivedData(HDDEDATA hData)
{
	DWORD len;

	if( hData )
	{
		len = DdeGetData( hData, NULL, 0, 0);
		if( RetData )
		{
        	// User-created string: check for overflow
			if( len > LenRetData )
				strcpy(errortext, "data overflow");
		}
		else
		{
            // Dynamically allocate return string
			RetData = new char[len];
			LenRetData = len;
        }
		DdeGetData( hData, RetData, LenRetData, 0);
    }
}


// This is the static callback funtion
HDDEDATA FAR PASCAL _export DDEManager::CallBack(WORD Type, WORD,
	HCONV Conv, HSZ, HSZ, HDDEDATA Data, DWORD, DWORD)
{
	// Static this pointer must be set
	if( !SThis )
		return NULL;

	switch( Type )
    {
		case XTYP_ADVDATA:
			if( Conv == SThis->hConv )
				SThis->ReceivedData(Data);
			return (HDDEDATA) DDE_FACK;

		case XTYP_XACT_COMPLETE:
			if( Conv == SThis->hConv )
				SThis->ReceivedData(Data);
			break;

		case XTYP_ERROR:
			if( Conv == SThis->hConv )
				strcpy(SThis->errortext, "critical DDE error");

		case XTYP_DISCONNECT:
			if( Conv == SThis->hConv )
				SThis->Disconnect();

		default:
			break;
    }
            
	return NULL;
}



